#if 0
#   Copyright (C) 2010 by William A. Lorenz
#   Program files for software "RNAlocopt" under GNU public license,
#   details of which can be found at:
#       http://www.gnu.org/licenses/gpl-3.0.txt
#
#   Ideas developed by collaboration of W.A. Lorenz and P.G. Clote
#   and funded by NSF Grant DBI-0543506 to PGC.
#   Any opinions, findings, and conclusions or recommendations expressed 
#   in this material are those of the authors and do not necessarily 
#   reflect the views of the National Science Foundation.
#
#   If you use this software, we would be grateful if you could cite the
#   paper: "Computing the partition function for kinetically trapped RNA
#   secondary structures", by W.A. Lorenz and P. Clote, PLoS ONE (2010).
#
#   Algorithm and code developed by W.A. Lorenz, with some code concerning
#   Turner energy model taken from Vienna RNA Package, developed by  
#   IL Hofacker et al. http://www.tbi.univie.ac.at/RNA/, which is under
#   GNU public license as well.
#
#   This file is free software; you can redistribute it and/or modify it
#   under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#  
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   General Public License for more details.
#  
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA
#   02110-1301, USA.
#  
#endif




/* Program to calculate and store irred HP's */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "convert_Vienna.h"
#include "pair_mat.h"

#define EPSILON 0.00001
#define PRIVATE static
#define MIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))

PRIVATE double **HP_energy_part;
PRIVATE double part, energy;
PRIVATE int is_min(int i, int j, short *S0, char *sequence, double kT);

void find_irred_HP(short *S0, char *sequence, double kT, int n){
  int i,j;
  HP_energy_part=(double**) malloc((n+1)*sizeof(double*));
  for (i=1;i<=n;i++)
    HP_energy_part[i]=(double *) calloc(n+1, sizeof(double));
  for (i=1;i<=n;i++)
    for (j=i+4;j<=n;j++)
      if (pair[S0[i]][S0[j]]){
	if (is_min(i,j, S0, sequence, kT)){
	  HP_energy_part[j][i]=part;
	}
	HP_energy_part[i][j]=energy;
      }
  return;
}
double HP_E(int i, int j){
  return HP_energy_part[i][j];
}
double HP_part(int i, int j){
  return HP_energy_part[j][i];
}
double HP_is_min(int i,int j){
  return HP_energy_part[j][i];
}
PRIVATE int is_min(int i, int j, short *S0, char *sequence, double kT){
  int ip, jp;
  
  energy=HP_Energy(i,j,S0, sequence);
  for (ip=i+1;ip<=MIN(i+30,j-1);ip++)
    for (jp=j-1;jp>=MAX(j-(30-(ip-i)),ip+4);jp--)
      if (pair[S0[ip]][S0[jp]]){
	if (IL_Energy(i,j,ip,jp, S0)+HP_Energy(ip,jp, S0,sequence)
	    +EPSILON<energy){
	  return 0;
	}
      }
  part=exp(-energy/kT);
  return 1;
}

  
