#if 0
#   Copyright (C) 2010 by William A. Lorenz
#   Program files for software "RNAlocopt" under GNU public license,
#   details of which can be found at:
#       http://www.gnu.org/licenses/gpl-3.0.txt
#
#   Ideas developed by collaboration of W.A. Lorenz and P.G. Clote
#   and funded by NSF Grant DBI-0543506 to PGC.
#   Any opinions, findings, and conclusions or recommendations expressed 
#   in this material are those of the authors and do not necessarily 
#   reflect the views of the National Science Foundation.
#
#   If you use this software, we would be grateful if you could cite the
#   paper: "Computing the partition function for kinetically trapped RNA
#   secondary structures", by W.A. Lorenz and P. Clote, PLoS ONE (2010).
#
#   Algorithm and code developed by W.A. Lorenz, with some code concerning
#   Turner energy model taken from Vienna RNA Package, developed by  
#   IL Hofacker et al. http://www.tbi.univie.ac.at/RNA/, which is under
#   GNU public license as well.
#
#   This file is free software; you can redistribute it and/or modify it
#   under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#  
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   General Public License for more details.
#  
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA
#   02110-1301, USA.
#  
#endif




/* module that stores IL data */
#include <stdio.h>
#include <stdlib.h>
#include "IL.h"

#define PRIVATE static
#define SIZE_INIT 6

PRIVATE int n, i_last=0, j_last=0, index=0;
PRIVATE int **IL_table;
PRIVATE int pres_size=SIZE_INIT, size_inc=2;
PRIVATE void chk_alloc();

void IL_initialize(int n_in){
  int i;

  n=n_in;
  IL_table=(int**) malloc((n+1)*sizeof(int*));
  for (i=1;i<=n;i++)
    IL_table[i]=(int *) calloc(n+1, sizeof(int));
  IL = (struct IL_info *) malloc (pres_size*n*n*sizeof(struct IL_info));
  return;
}

void IL_add(double part, double E0, int i, int j, int ip, int jp, double count){
  if (i==i_last && j==j_last){
    index++; IL_table[j][i]++;
  } else {
    index++;IL_table[i][j]=index;IL_table[j][i]=1;
    i_last=i;j_last=j;
  }
  chk_alloc();
  IL[index].ip=ip;IL[index].jp=jp;IL[index].part=part;
  IL[index].energy=E0;IL[index].count=count;
  return;
}  
int IL_start(int i, int j){
  return IL_table[i][j];
}
int IL_end(int i, int j){
  return IL_table[i][j]+IL_table[j][i]-1;
}
int IL_index(){
  return index;
}

PRIVATE void chk_alloc(){
  if (index>=pres_size*n*n){
    pres_size+=size_inc;
    if (realloc(IL, pres_size*n*n*sizeof(struct IL_info))==NULL){
      printf("failed memory allocation.\n");
      exit(EXIT_FAILURE);
    }
  }
  return;
}
