#if 0
#   Copyright (C) 2010 by William A. Lorenz
#   Program files for software "RNAlocopt" under GNU public license,
#   details of which can be found at:
#       http://www.gnu.org/licenses/gpl-3.0.txt
#
#   Ideas developed by collaboration of W.A. Lorenz and P.G. Clote
#   and funded by NSF Grant DBI-0543506 to PGC.
#   Any opinions, findings, and conclusions or recommendations expressed 
#   in this material are those of the authors and do not necessarily 
#   reflect the views of the National Science Foundation.
#
#   If you use this software, we would be grateful if you could cite the
#   paper: "Computing the partition function for kinetically trapped RNA
#   secondary structures", by W.A. Lorenz and P. Clote, PLoS ONE (2010).
#
#   Algorithm and code developed by W.A. Lorenz, with some code concerning
#   Turner energy model taken from Vienna RNA Package, developed by  
#   IL Hofacker et al. http://www.tbi.univie.ac.at/RNA/, which is under
#   GNU public license as well.
#
#   This file is free software; you can redistribute it and/or modify it
#   under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#  
#   This program is distributed in the hope that it will be useful, but
#   WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#   General Public License for more details.
#  
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA
#   02110-1301, USA.
#  
#endif




/* Program to convert from my standard notation for energy of IL's, HP's and
ML's from existing Vienna package subroutines. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "fold.h"
#include "energy_const.h"
#include "fold_vars.h"
#include "pair_mat.h"
#include "convert_Vienna.h"
#include "params.h"

#define PRIVATE static

void Initialize_Params(){
  P = scale_parameters();//from params.c, gets our parameters, for a given temp
}

float HP_Energy(int i, int j, short *S0, char* sequence){
  int type, energy_int;

  type=pair[S0[i]][S0[j]];
  energy_int=HairpinE(j-i-1, type, S0[i+1], S0[j-1], sequence+i-1);
  //  printf("i,j,HP_energy are %d,%d,%d.\n", i, j, energy_int);
  return ((float) energy_int)/100.;
}
float IL_Energy(int i, int j, int ip, int jp, short *S0){
  int type, type2, energy_int, k, l;

  type=pair[S0[i]][S0[j]];
  type2=pair[S0[jp]][S0[ip]];
  energy_int=LoopEnergy(ip-i-1, j-jp-1, type, type2,
			S0[i+1], S0[j-1], S0[ip-1], S0[jp+1]);
  //  printf ("energy_int is now %d\n", energy_int);
  //getchar();
  return ((float) energy_int)/100.;
}  
float EL_Energy(int i, int j, short *S0){
  int type, energy_int;

  type=pair[S0[i]][S0[j]];
  energy_int=P->MLintern[type]-P->MLintern[1]; /* 0 or AU penalty */
  return (float) energy_int/100.;
  //return 3.5;
}
float ML_Energy(int i, int j, short *S0){
  int type, energy_int;
  int k;
  
  type=pair[S0[i]][S0[j]];
  energy_int=P->MLintern[type]; 
  //  printf("i,j,EL_energy are %d,%d,%d.\n",i,j,energy_int);
  return (float) energy_int/100.;
}
			
  
